#!/usr/bin/python
#############################################################################
# Name:         utility.py
# Purpose:      A configurable SSH agent utility for ZOPE3.
#               Utility classes
# Maintainer:   Torsten Kurbad <t.kurbad@iwm-kmrc.de>
# Copyright:    (c) iwm-kmrc.de KMRC - Knowledge Media Research Center
# License:      GPLv2
#############################################################################
__docformat__ = 'restructuredtext'

import os
import re
import signal
from subprocess import call, Popen, PIPE

from persistent import Persistent
from persistent.list import PersistentList
from zope.interface import implements
from zope.component import getSiteManager
from zope.component import queryUtility
from zope.schema.fieldproperty import FieldProperty

from zope.app.component.vocabulary import UtilityVocabulary
from zope.app.container.contained import Contained

from sshagent.interfaces import ISSHAgent
from sshagent.interfaces import (SSHAgentNotRunning,
                                 IdentityListEmpty,
                                 NotAllIdentitiesAdded,
                                 NotAllIdentitiesRemoved)


class SSHAgent(Persistent, Contained):
    """A SSHAgent utility.
    
    Provides methods to handle SSH private keys stored, e.g. in the ZODB.

    """

    implements(ISSHAgent)

    ssh_agent = FieldProperty(ISSHAgent['ssh_agent'])
    ssh_add = FieldProperty(ISSHAgent['ssh_add'])

    _autoRegister = True

    def __init__(self, keys=None):
        self.env = os.environ
        self.sock = None
        self.pid = None
        self._identities = PersistentList()

    def _unique(self, list):
        """Return list with all duplicate items removed."""
        result = []
        for item in list:
            if item not in result:
                result.append(item)
        return result

    def getAutoRegister(self):
        """Return autoRegister status."""
        return self._autoRegister

    def setAutoRegister(self, value):
        """Register the utility automagically as 'SSHAgent'."""
        if value:
            siteManager = getSiteManager(self)

            if queryUtility(ISSHAgent, name='SSHAgent') is None:
                siteManager.registerUtility(self, ISSHAgent, 'SSHAgent')

        self._autoRegister = value

    autoRegister = property(getAutoRegister, setAutoRegister)

    def getIdentities(self):
        """Getter for attribute 'identities'."""
        return self._identities

    def setIdentities(self, values):
        """Setter for attribute 'identities'."""
        self._identities = self._unique(values)

    identities = property(getIdentities, setIdentities)

    def _parseEnvVar(self, eval, var):
        """Extract the environment variables from the output of ssh-agent."""
        regex = re.compile('%s=.*?;' % var)
        resultlist = regex.findall(eval)
        if resultlist:
            result = resultlist[0].rstrip(';').split('=')[1]
            return result

    def callAgent(self):
        """Call the requested SSH agent command and read it's environment."""
        if self.pid is not None or self.sock is not None:
            return

        process = Popen(self.ssh_agent.encode('utf-8'),
                        stdout=PIPE,
                        stderr=PIPE)

        output, errors = process.communicate()

        self.sock = self._parseEnvVar(output, 'SSH_AUTH_SOCK')
        self.pid = self._parseEnvVar(output, 'SSH_AGENT_PID')
        if self.sock is None or self.pid is None:
            raise SSHAgentNotRunning

        self.env['SSH_AUTH_SOCK'] = self.sock
        self.env['SSH_AGENT_PID'] = self.pid

    def killAgent(self):
        """Terminate a previously called SSH agent. """
        if self.pid is None:
            return

        os.kill(int(self.pid), signal.SIGKILL)
        self.sock = None
        self.pid = None

    def addIdentities(self):
        """Add identies to the currently running agent via the configured SSH
        add command.

        """
        if self.identities == []:
            raise IdentityListEmpty
        if self.pid is None or self.sock is None:
            raise SSHAgentNotRunning

        command = [self.ssh_add.encode('utf-8')]

        def_umask = os.umask(0177)
        filenames = []
        for identity in self.identities:
            filename = os.tmpnam()
            fd = open(filename, 'w+b')
            fd.write(identity)
            fd.close()
            filenames.append(filename)
        os.umask(def_umask)

        command.extend(filenames)

        process = Popen(command,
                        stdout=PIPE,
                        stderr=PIPE,
                        env=self.env)

        stdout, stderr = process.communicate()

        for filename in filenames:
            os.unlink(filename)

        num_added = len(re.compile('Identity added:').findall(stderr))
        num_identities = len(self.identities)
        if num_added != num_identities:
            raise NotAllIdentitiesAdded
        return num_added

    def clearIdentities(self):
        """Remove all identies from the currently running agent."""
        command = [self.ssh_add.encode('utf-8'), '-D']

        process = Popen(command, stdout=PIPE, stderr=PIPE, env=self.env)

        stdout, stderr = process.communicate()

        if not 'All identities removed.' in stderr:
            raise NotAllIdentitiesRemoved

    def getEnv(self):
        """Return the utilities os.environ."""
        return self.env


class SSHAgentVocabulary(UtilityVocabulary):
    interface = ISSHAgent
    nameOnly = True
